#!/usr/bin/python
#Filename: panelwork.py
'''panelwork.py is a tool for drawing front/rear panels of electronic equipment.

   It generates PostScript which can easily be printed at the correct (real units) scale.
   Printing could be done via GhostScript / GhostView for example.
   
   Either the drill patterns or front panel printing (or both) can be output.
   
   Each component is defined by a pair of procedures: one draws the drill pattern, the other
   draws either the drill pattern or the front panel legend for the component.
   
   All dimensions are in tenths of a millimeter.

   It may be an idea to print on to adhesive labels. Possibly useful labels are:
   Avery L7784-25 A4 Transparent Laser
   Avery L4775-20 A4 Heavy Duty White Laser
   Avery L6013-20 A4 Silver Heavy Duty Laser
   The result may be protected with spray-on acrylic varnish. E.g:
   Gerstaecker Acrylic Spray Varnish (Matt/Semi-Matt/Gloss)
'''

import math

#
# Global variables ...
#

dolabels = False
domarks = True
docomponents = True
numpages = 0
curpage = 0
doxmirror = 0
xmirrorpivot = 0

#
# Utility functions ...
#

def stripInnerZeros( s ):
    '''Remove the first set of zeros in string s, possibly following other characters.'''
    zfirst = s.find( '0' )
    if( zfirst >= 0 ):
        return( s[:zfirst] + s[zfirst:].lstrip( '0' ) )

def addSpacesToColon( s, n=8 ):
    '''Add spaces so that the number of characters before the first colon in s is n.'''
    cfirst = s.find( ':' )
    if( cfirst > 0 ):
        return( s[:cfirst].rjust( n ) + s[cfirst:] )

#
# Select what to draw ...
#

def doDrillMarks( yes ):
    '''Draw drill marks. These indicate where to drill holes of various sizes for the components. Also accumulate component lists.'''
    global domarks
    domarks = yes

def doControlLabels( yes ):
    '''Draw front panel labels for the components. This outputs teh graphics for the front panel.'''
    global dolabels
    dolabels = yes

def doComponentRefs( yes ):
    '''Accumulate component references when drawing drill marks.'''
    global docomponents
    docomponents = yes

def doXMirror( yes, xpivot ):
    '''Mirror component positions about xpivot.'''
    global doxmirror
    global xmirrorpivot
    doxmirror = yes
    xmirrorpivot = xpivot

#
# Basic elements ...
#

def createPSFile( name ):
    '''Create a PostScript output file.'''
    try:
        f = open( name, 'w' );
    except IOError:
        print( 'Failed to create {0}'.format(name) )
    else:
        global numpages
        numpages = 0
        
        print( '%!PS-Adobe-3.0', file=f )
        print( '%%Pages: (atend)', file=f )
        print( '%%EndComments', file=f )
        print( '%%BeginProlog', file=f )
        print( '/l {moveto lineto currentpoint stroke moveto} def', file=f )
        print( '/m {newpath moveto} def', file=f )
        print( '/c {setrgbcolor} def', file=f )
        print( '/s {0.283465 mul} def', file=f )
        print( '/w {s setlinewidth} def', file=f )
        print( '/fsd {findfont exch scalefont def} bind def', file=f )
        print( '/sms {setfont moveto show} bind def', file=f )
        print( '/ms {moveto show} bind def', file=f )
        print( '/f1 7 /Helvetica fsd', file=f )
        print( '/f2 9 /Helvetica-Bold fsd', file=f )
        print( '/f3 4 /Helvetica fsd', file=f )
        print( '/fh { currentfont /FontMatrix get 3 get 1000 mul } bind def', file=f )
        print( '%%EndProlog', file=f )
        return f

def endPSFile( f ):
    '''End and close a PostScript output file.'''
    global numpages
    print( '%%Trailer', file=f )
    print( '%%Pages: {0}'.format( numpages ), file=f )
    print( '%%EOF', file=f )
    f.close()

def beginPage( f ):
    '''Start a new page in  a PostScript output file.'''
    global numpages
    numpages += 1
    print( '%%Page: {0} {0}'.format( numpages ), file=f )

def endPage( f ):
    '''End the current page of the PostScript output file.'''
    print( 'showpage', file=f )

def getPageCount():
    return numpages
    
def shift( f, dx, dy ):
    '''Offset subsequent PostScript output by (dx,dy) wrt page.'''
    print( 'gsave {0} s {1} s translate'.format( dx, dy ), file=f )

def endshift( f ):
    '''Remove offset wrt page from subsequent PostScript output.'''
    print( 'grestore', file=f )

def beginColour( f, r, g, b ):
    '''Start drawing in the specified colour. Default in black.'''
    print( 'gsave {0} {1} {2} setrgbcolor'.format( r, g, b ), file=f )

def endColour( f ):
    '''Stop drawing in colour.'''
    print( 'grestore', file=f )

def xM( x ):
    '''If X mirror is required, mirror in X.'''
    if( doxmirror ):
        return 2 * xmirrorpivot - x
    else:
        return x
    
def drawLine( f, x1, y1, x2, y2 ):
    '''Draw a solid line from (x1,y1) to (x2,y2)'''
    print( '{0} s {1} s {2} s {3} s l'.format( xM(x1), y1, xM(x2), y2 ), file=f )

def drawLineDotted( f, x1, y1, x2, y2 ):
    '''Draw a dotted line from (x1,y1) to (x2,y2)'''
    print( '[2] 1 setdash', file=f )
    drawLine( f, x1, y1, x2, y2 )
    print( '[] 0 setdash', file=f )

def drawLineJoiningCircles( f, x1, y1, r1, x2, y2, r2 ):
    '''Draw a line joining circle 1 (x1,y1), radius r1 to circle 2 (x2,y2) radius r2.'''
    dx = x2 - x1
    dy = y2 - y1
    length = math.sqrt( dx*dx + dy*dy )
    if( length < 1 ):
        drawLine( f, x1, y1, x2, y2 )
    else:
        dx /= length
        dy /= length
        x3 = x1 + r1 * dx
        y3 = y1 + r1 * dy
        x4 = x2 - r2 * dx
        y4 = y2 - r2 * dy
        drawLine( f, x3, y3, x4, y4 )

def drawDogLegJoiningCircles( f, x1, y1, r1, x2, y2, r2 ):
    '''Draw a bent line joining circle 1 (x1,y1), radius r1 to circle 2 (x2,y2) radius r2.'''
    xm = 0.5 * ( x1 + x2 )
    drawLineJoiningCircles( f, x1, y1, r1, xm, y2, 0.0 )
    drawLineJoiningCircles( f, xm, y2, 0.0, x2, y2, r2 )

def drawCircle( f, x, y, r ):
    '''Draw a solid circle, center (x,y) radius r.'''
    print( '{0} s {1} s moveto'.format( xM(x)+r, y ), file=f )
    print( '{0} s {1} s {2} s 0 360 arc currentpoint stroke moveto'.format( xM(x), y, r ), file=f )

def drawCircleDotted( f, x, y, r ):
    '''Draw a dotted circle, center (x,y) radius r.'''
    print( '[2] 1 setdash', file=f )
    drawCircle( f, x, y, r )
    print( '[] 0 setdash', file=f )

def drawRect( f, x, y, w, h ):
    '''Draw a rectangle, center at (x,y), width w, height h.'''
    drawLine( f, x-w*0.5, y-h*0.5, x+w*0.5, y-h*0.5 )
    drawLine( f, x+w*0.5, y-h*0.5, x+w*0.5, y+h*0.5 )
    drawLine( f, x+w*0.5, y+h*0.5, x-w*0.5, y+h*0.5 )
    drawLine( f, x-w*0.5, y+h*0.5, x-w*0.5, y-h*0.5 )

def drawArc( f, x, y, r, a1, a2 ):
    '''Draw an arc, center at (x,y), radius r, from angle a1 to angle a2.'''
    print( '{0} s {1} s moveto'.format( xM(x)+r*math.cos( math.radians(a1) ), y+r*math.sin( math.radians(a1) ) ), file=f )
    print( '{0} s {1} s {2} s {3} {4} arc currentpoint stroke moveto'.format( xM(x), y, r, a1, a2 ), file=f )

def drawRectRounded( f, x, y, w, h, r ):
    '''Draw a rectangle, center at (x,y), width w, height h, with rounded corners, radius r.'''
    drawLine( f, x-w*0.5+r, y-h*0.5, x+w*0.5-r, y-h*0.5 )
    drawLine( f, x+w*0.5, y-h*0.5+r, x+w*0.5, y+h*0.5-r )
    drawLine( f, x+w*0.5-r, y+h*0.5, x-w*0.5+r, y+h*0.5 )
    drawLine( f, x-w*0.5, y+h*0.5-r, x-w*0.5, y-h*0.5+r )
    if( doxmirror ):
        drawArc( f, x-w*0.5+r, y+h*0.5-r, r, 0, 90 )
        drawArc( f, x+w*0.5-r, y+h*0.5-r, r, 90, 180 )
        drawArc( f, x+w*0.5-r, y-h*0.5+r, r, 180, 270 )
        drawArc( f, x-w*0.5+r, y-h*0.5+r, r, 270, 360 )
    else:
        drawArc( f, x+w*0.5-r, y+h*0.5-r, r, 0, 90 )
        drawArc( f, x-w*0.5+r, y+h*0.5-r, r, 90, 180 )
        drawArc( f, x-w*0.5+r, y-h*0.5+r, r, 180, 270 )
        drawArc( f, x+w*0.5-r, y-h*0.5+r, r, 270, 360 )

def drawWedgeRounded( f, x, y, wb, h, wt, r ):
    '''Draw a wedge, center at (x,y), bottom width wb, height h, top width wt, with rounded corners, radius r.'''
    drawLine( f, x-wb*0.5+r, y-h*0.5, x+wb*0.5-r, y-h*0.5 )
    drawLine( f, x+wb*0.5, y-h*0.5+r, x+wt*0.5, y+h*0.5-r )
    drawLine( f, x+wt*0.5-r, y+h*0.5, x-wt*0.5+r, y+h*0.5 )
    drawLine( f, x-wt*0.5, y+h*0.5-r, x-wb*0.5, y-h*0.5+r )
    if( doxmirror ):
        drawArc( f, x-wt*0.5+r, y+h*0.5-r, r, 0, 90 )
        drawArc( f, x+wt*0.5-r, y+h*0.5-r, r, 90, 180 )
        drawArc( f, x+wb*0.5-r, y-h*0.5+r, r, 180, 270 )
        drawArc( f, x-wb*0.5+r, y-h*0.5+r, r, 270, 360 )
    else:
        drawArc( f, x+wt*0.5-r, y+h*0.5-r, r, 0, 90 )
        drawArc( f, x-wt*0.5+r, y+h*0.5-r, r, 90, 180 )
        drawArc( f, x-wb*0.5+r, y-h*0.5+r, r, 180, 270 )
        drawArc( f, x+wb*0.5-r, y-h*0.5+r, r, 270, 360 )
        
def drawText( f, x, y, s, font='f1' ):
    '''Draw text string s, using font, left justified, starting at (x,y).'''
    if( doxmirror ):
        print( '{3} setfont {0} s {1} s moveto ({2}) stringwidth pop neg 0 rmoveto ({2}) show'.format( xM(x), y, s, font ), file=f )
    else:
        print( '({0}) {1} s {2} s {3} sms'.format( s, xM(x), y, font ), file=f )

def drawTextR( f, x, y, s, font='f1' ):
    '''Draw text string s, using font, right justified, ending at (x,y).'''
    if( doxmirror ):
        print( '({0}) {1} s {2} s {3} sms'.format( s, xM(x), y, font ), file=f )
    else:
        print( '{3} setfont {0} s {1} s moveto ({2}) stringwidth pop neg 0 rmoveto ({2}) show'.format( xM(x), y, s, font ), file=f )

def drawTextC( f, x, y, s, font='f1' ):
    '''Draw text string s, using font, centered in x at (x,y).'''
    print( '{3} setfont {0} s {1} s moveto ({2}) stringwidth pop neg 0.5 mul 0 rmoveto ({2}) show'.format( xM(x), y, s, font ), file=f )

def drawTextCC( f, x, y, s, font='f1' ):
    '''Draw text string s, using font, centered in x and y at (x,y).'''
    print( '{3} setfont {0} s {1} s moveto ({2}) stringwidth pop neg 0.5 mul fh neg 0.5 mul rmoveto ({2}) show'.format( xM(x), y, s, font ), file=f )

def drawTextCB( f, x, y, s, font='f1' ):
    '''Draw text string s, using font, centered at (x,y), shifted in y for use at the bottom of controls.'''
    print( '{3} setfont {0} s {1} s moveto ({2}) stringwidth pop neg 0.5 mul fh neg 2 add rmoveto ({2}) show'.format( xM(x), y, s, font ), file=f )

def drawEdges( f, x, y, w, h, title=' ' ):
    '''Draw edges with coordinate labels, bottom left (x,y), width w, height h.'''
    drawLine( f, x, y, x+w, y )
    drawLine( f, x+w, y, x+w, y+h )
    drawLine( f, x+w, y+h, x, y+h )
    drawLine( f, x, y+h, x, y )
    for i in range(x, x+w, 100):
        drawLine( f, i, y, i, y-20 )
        drawLine( f, i, y+h, i, y+h+20 )
        s = '{0}'.format( i )
        drawTextC(  f, i, y+h+30, s )
        drawTextC(  f, i, y-40, s )
    for i in range(y, y+h, 100 ):
        s = '{0}'.format( i )
        drawTextR( f, x-30, i-7, s )
        drawText( f, x+w+30, i-7, s )
        drawLine( f, x, i, x-20, i )
        drawLine( f, x+w, i, x+w+20, i )
    for i in range(x+50, x+w, 100):
        drawLine( f, i, y, i, y-10 )
        drawLine( f, i, y+h, i, y+h+10 )
    for i in range(y+50, y+h, 100 ):
        drawLine( f, x, i, x-10, i )
        drawLine( f, x+w, i, x+w+10, i )
    drawTextC( f, x+0.5*w, y+h+100, title, 'f2' )
    #drawLine( f, x+w/2, y, x+w/2, y+h )
        
def lineWidth( f, w ):
    '''Set the width of the pen with which to draw lines.'''
    print( '{0} w'.format( w ), file=f )

def grayBox( f, x, y, w, h, gray ):
    '''Draw a filled rectangle, bottom left (x,y), width w, height h, in gray level gray (0=black, 1=white).'''
    print( 'gsave newpath {0} s {1} s moveto'.format( xM(x), y ), file=f )
    print( '{0} s {1} s lineto'.format( xM(x+w), y ), file=f )
    print( '{0} s {1} s lineto'.format( xM(x+w), y+h ), file=f )
    print( '{0} s {1} s lineto'.format( xM(x), y+h ), file=f )
    print( 'closepath {0} setgray fill grestore'.format( gray ), file=f )

def colourBox( f, x, y, w, h, r, g, b ):
    '''Draw a filled rectangle, bottom left (x,y), width w, height h, in RGB colour r, g, b (0 to 1).'''
    print( 'gsave newpath {0} s {1} s moveto'.format( xM(x), y ), file=f )
    print( '{0} s {1} s lineto'.format( xM(x+w), y ), file=f )
    print( '{0} s {1} s lineto'.format( xM(x+w), y+h ), file=f )
    print( '{0} s {1} s lineto'.format( xM(x), y+h ), file=f )
    print( 'closepath {0} {1} {2} setrgbcolor fill grestore'.format( r, g, b ), file=f )

def drillMark( f, x, y, d, alt=False ):
    '''Draw a drill mark for a circular hole, hole center (x,y), diameter d.'''
    r = 0.5 * d
    lineWidth( f, 1 )
    if( alt ):
        beginColour( f, 0, 1, 0 )
    drawCircle( f, x, y, r )
    drawLine( f, x-r-10, y, x+r+10, y )
    drawLine( f, x, y-r-10, x, y+r+10 )
    s = '{0:.1f}'.format(0.1*d)
    if( alt ):
        drawTextR( f, x-5-0.5*d, y+5, s )
    else:
        drawText( f, x+5, y+5, s )
    if( alt ):
        endColour( f )

def drillRect( f, x, y, w, h, dd=40 ):
    '''Draw a drill pattern for a rectangular hole, center (x,y) width w, height h. Use a drill bit diameter dd.'''
    drawRect( f, x, y, w, h )
    # If two holes will not fit on an edge, reduce the hole size so they will,
    # rounding down to the nearest 0.5 mm. If the size ends up at zero, give up.
    dd2 = 0.5 * dd
    if( w < ( 2 * dd ) ):
        dd = 0.5 * w
        dd = 5 * int( dd / 5 )
    if( h < ( 2 * dd ) ):
        dd = 0.5 * h
        dd = 5 * int( dd / 5 )
    if( dd == 0 ):
        return
    # Holes at the corners.
    drillMark( f, x-0.5*w+dd2, y-0.5*h+dd2, dd )
    drillMark( f, x+0.5*w-dd2, y-0.5*h+dd2, dd )
    drillMark( f, x+0.5*w-dd2, y+0.5*h-dd2, dd )
    drillMark( f, x-0.5*w+dd2, y+0.5*h-dd2, dd )
    # Find number of holes along the edges.
    nholew = w // dd
    nholeh = h // dd
    # Find the step between hole centers.
    dw = dd + float( w - nholew * dd ) / ( nholew - 1 )
    dh = dd + float( h - nholeh * dd ) / ( nholeh - 1 )
    # Holes along the edges.
    for i in range(1, nholew-1):
        drillMark( f, x-0.5*w+dd2+i*dw, y-0.5*h+dd2, dd )
        drillMark( f, x-0.5*w+dd2+i*dw, y+0.5*h-dd2, dd )
    for i in range(1, nholeh-1):
        drillMark( f, x-0.5*w+dd2, y-0.5*h+dd2+i*dh, dd )
        drillMark( f, x+0.5*w-dd2, y-0.5*h+dd2+i*dh, dd )
    # Label.
    s = '{0:.1f} x {1:.1f}'.format(0.1*w, 0.1*h)
    drawTextC( f, x, y, s )

def drillWedgeRounded( f, x, y, wb, h, wt, r ):
    drawWedgeRounded( f, x, y, wb, h, wt, r )
    drillMark( f, x-0.5*wb+r, y-0.5*h+r, 2*r )
    drillMark( f, x+0.5*wb-r, y-0.5*h+r, 2*r )
    drillMark( f, x-0.5*wt+r, y+0.5*h-r, 2*r )
    drillMark( f, x+0.5*wt-r, y+0.5*h-r, 2*r )

def outlineShifted( f, x, y, w, h, dx, dy ):
    '''Draw a component outline, as a rectangle, center (x,y), width w, height h, offset from center (dx,dy).'''
    drawLineDotted( f, x-0.5*w+dx, y-0.5*h+dy, x+0.5*w+dx, y-0.5*h+dy )
    drawLineDotted( f, x+0.5*w+dx, y-0.5*h+dy, x+0.5*w+dx, y+0.5*h+dy )
    drawLineDotted( f, x+0.5*w+dx, y+0.5*h+dy, x-0.5*w+dx, y+0.5*h+dy )
    drawLineDotted( f, x-0.5*w+dx, y+0.5*h+dy, x-0.5*w+dx, y-0.5*h+dy )    

def outline( f, x, y, w, h ):
    '''Draw a component outline, as a rectangle, center (x,y), width w, height h.'''
    outlineShifted( f, x, y, w, h, 0, 0 )

def outlineLabel( f, x, y, w, h, s, c=None ):
    '''As outline(), labelled with string s. If c is not None, use it as a component reference.'''
    outline( f, x, y, w, h )
    drawTextR( f, x+0.5*w, y-0.5*h, s )
    if( docomponents ):
        if c is not None:
            drawText( f, x-0.5*w, y+0.5*h+3, stripInnerZeros(c), 'f3' )

def outlineLabelShifted( f, x, y, w, h, dx, dy, s, c=None ):
    '''As outlineShifted(), labelled with string s.'''
    outlineShifted( f, x, y, w, h, dx, dy )
    drawTextR( f, x+0.5*w+dx, y-0.5*h+dy, s )
    if( docomponents ):
        if c is not None:
            drawText( f, x-0.5*w+dx, y+0.5*h+dy+3, stripInnerZeros(c), 'f3' )

#
# Components ...
#
# To add a new component:
# 1. Think of its name (e.g. xlr)
# 2. Add its name to the compCounts array.
# 3. Decide what type it is (e.g. diode=D, connector=J ... )
# 4. Associate it with its type in the compTypes array.
# 5. If you have invented a new type for it, add that type to the typeCounts array.
# 6. Create a routine for it that draws it drill pattern. Call it by its name, E.g. xlr().
#     - This routine must indicate where holes, etc., are to be drilled.
#     - If the component is not symmetrical in X, you may have to account for mirroring in X.
#     - Bump compCounts[] for the component.
#     - Draw the outlineLabel() which indicates the area taken up the component.
#        - Call compRefString() while doing this to correctly update all the component lists.
# 7. Create a routine that draw the front panel graphics fir the component. Call it <name>Label.
#   e.g. xlrLabel(). 
#

# List of defined components and counts of each.
compCounts = { 'led':0, \
               'sw1':0, \
               'xlr':0, \
               'phono':0, \
               'pot':0, \
               'fuse':0, \
               'inlet':0, \
               'swmains':0, \
               'jackqt':0, \
               'screw':0, \
               'd9pin':0 \
               }

# Relate defined component to component type.
compTypes = { 'led':'D', \
             'sw1':'SW', \
             'xlr':'J', \
             'phono':'J', \
             'pot':'VR', \
             'fuse':'F', \
             'inlet':'J', \
             'swmains':'SW', \
             'jackqt':'J', \
             'screw':'S', \
             'd9pin':'J' \
             }

# Counts of each component type.
typeCounts = { 'D':0, \
               'SW':0, \
               'J':0, \
               'VR':0, \
               'F':0, \
               'S':0 \
               }

# List of components.
compList = []

def compRefString( cName, desc=' ' ):
    '''Given a component name and optional description, keep a count of component types, and generate a list of components.'''
    typeCounts[ compTypes[cName] ] += 1
    # Make a component reference from type and count. Format the count with leading zeros for easy sorting.
    cref = '{0}{1:0>3}'.format( compTypes[cName], typeCounts[ compTypes[cName] ] )
    # Add the component reference with description to the list of components.
    listEntry = '{0:>6} : {1}'.format( cref, desc )
    compList.append( listEntry )
    return cref

def led( f, x, y ):
    '''3mm LED drill pattern.

    Avago HLMP-1301 (red)    [Farnell 100 3196],
              -1401 (yellow) [Farnell 100 3198],
              -1503 (green). [Farnell 100 3199] '''
    desc = 'LED: Avago HLMP-1301 (red)    [Farnell 100 3196]'
    drillMark( f, x, y, 30 )
    compCounts['led'] += 1
    outlineLabel( f, x, y, 35, 35, 'led', compRefString('led', desc) )

def ledLabel( f, x, y, label=' ' ):
    '''3mm LED, panel or drill pattern. Avago HLMP-1301 (red), -1401 (yellow), -1503 (green).'''
    if( dolabels ):
        lineWidth( f, 2 )
        drawCircle( f, x, y, 35*0.5 )
        drawTextCB( f, x, y-37, label )
    if( domarks ):
        led( f, x, y )

def sw1( f, x, y ):
    '''Toggle switch drill pattern, mini, Multicomp 1MD.

    DPDT,            2 pos: 1MD1 T1B1M1QE [Farnell 947 3513]
    DPDT,  2 pos momentary: 1MD2 T1B5M1QE [Farnell 155 0181]
    DPDT, 3 pos center off: 1MD3 T1B1M1QE [Farnell 947 3530]'''
    desc = 'Toggle switch: Mini, Multicomp 1MD1 [Farnell 947 3513], 1MD3 [Farnell 947 3530]'
    drillMark( f, x, y, 65 )
    drillMark( f, x, y+61, 25 )
    compCounts['sw1'] += 1
    outlineLabel( f, x, y, 115, 127, 'sw1', compRefString('sw1', desc) )

def sw1Label( f, x, y, bot, top, mid = ' ', rmid = False ):
    '''Toggle switch, panel or drill pattern, mini, Multicomp 1MD.'''
    if( dolabels ):
        lineWidth( f, 2 )
        drawCircle( f, x, y, 65*0.5 )
        drawTextCB( f, x, y-42, bot )
        drawTextC( f, x, y+42, top )
        if( mid != ' ' ):
            if( rmid ):
                drawText( f, x+37, y-6, mid )
            else:
                drawTextR( f, x-45, y-6, mid )
    if( domarks ):
        sw1( f, x, y )

def sw1LabelLedsR( f, x, y, bot, top, mid = ' ', rmid = False ):
    '''Toggle switch, Multicomp 1MD,  with tell-tale leds on the right.'''
    sw1Label( f, x, y, bot, top, mid, rmid )
    ledLabel( f, x+100, y-30 )
    ledLabel( f, x+100, y+30 )
    if( dolabels ):
        drawDogLegJoiningCircles( f, x, y, 65*0.5, x+100, y-30, 30*0.5 )
        drawDogLegJoiningCircles( f, x, y, 65*0.5, x+100, y+30, 30*0.5 )

def sw1LabelLedsL( f, x, y, bot, top, mid = ' ', rmid = False ):
    '''Toggle switch, Multicomp 1MD, with tell-tale leds on the left.'''
    sw1Label( f, x, y, bot, top, mid, rmid )
    ledLabel( f, x-100, y-30 )
    ledLabel( f, x-100, y+30 )
    if( dolabels ):
        drawDogLegJoiningCircles( f, x, y, 65*0.5, x-100, y-30, 30*0.5 )
        drawDogLegJoiningCircles( f, x, y, 65*0.5, x-100, y+30, 30*0.5 )
    
def sw1Sideways( f, x, y ):
    '''Toggle switch. Multicomp 1MD, on its side, drill pattern.'''
    desc = 'Toggle switch: Mini, Multicomp 1MD1 [Farnell 947 3513], 1MD3 [Farnell 947 3530]'
    drillMark( f, x, y, 65 )
    drillMark( f, x-61, y, 25 )
    compCounts['sw1'] += 1
    outlineLabel( f, x, y, 127, 115, 'sw1s', compRefString('sw1', desc) )

def sw1SidewaysLabel( f, x, y, left, right, mid=' ' ):
    '''Toggle switch. Multicomp 1MD, on its side, panel or drill pattern.'''
    if( dolabels ):
        lineWidth( f, 2 )
        drawCircle( f, x, y, 65*-.5 )
        drawTextCB( f, x-45, y-37, left )
        drawTextCB( f, x+45, y-37, right )
        if( mid != ' ' ):
            drawTextC( f, x, y+42, mid )
    if( domarks ):
        sw1Sideways( f, x, y )
    
def xlr( f, x, y ):
    '''XLR panel mounting, 3 pole, solder tag, Neutrik NC3 range, drill pattern.

    Female: NC3FDL-1, Farnell 251 021
    Male:   NC3MDL-1, Farnell 250 995
    Attach with screws: M3x8 countersunk 3mm dia, 8mm long, Farnell 1420 542
    and M3 full nut, Farnell 1419 447'''
    desc = 'XLR: Female: NC3FDL-1 [Farnell 251 021], Male:   NC3MDL-1 [Farnell 250 995]'
    drillMark( f, x, y, 240 )
    if doxmirror:
        drillMark( f, x-95, y-120, 35 )
        drillMark( f, x+95, y+120, 35 )
    else:
        drillMark( f, x+95, y-120, 35 )
        drillMark( f, x-95, y+120, 35 )
    compCounts['xlr'] += 1
    outlineLabel( f, x, y, 260, 310, 'xlr', compRefString('xlr', desc) )

def xlrLabel( f, x, y, label ):
    '''XLR panel mounting, panel or drill pattern.'''
    if( dolabels ):
        lineWidth( f, 2 )
        drawRectRounded( f, x, y, 280, 330, 30 )
        drawTextCB( f, x, y-0.5*310-20, label )
    if( domarks ):
        xlr( f, x, y )

def phono( f, x, y ):
    '''Phono socket, Deltron 431 (nickel), 432 (gold), drill pattern.

    Red,   Brass: Farnell 119 3844,  Gold: Farnell 119 3877
    Black, Brass: Farnell 119 3803,  Gold: Farnell 119 3876
    Washers, Isolating: Farnell 119 4026, Colour coding: Farnell 582 578 '''
    desc = 'Phono: Red, Brass: [Farnell 119 3844], Black, Brass: [Farnell 119 3803]'
    drillMark( f, x, y, 90 )
    compCounts['phono'] += 1
    outlineLabel( f, x, y, 95, 95, 'phono', compRefString('phono', desc) )

def phonoLabel( f, x, y, label ):
    '''Phono socket, label or drill pattern.'''
    if( dolabels ):
        lineWidth( f, 2 )
        drawCircle( f, x, y, 100*0.5 )
        drawTextCB( f, x, y-120*0.5, label )
    if( domarks ):
        phono( f, x, y )

def pot( f, x, y ):
    '''Potentiometer, rotary, Vishay Sfernice P11, drill pattern.

    Linear, Cermet, Single gang: Farnell 1141 595 to 605 (100,500,1K,5K,10K,50K,100K,500K,1M)
    Linear, Cermet, Dual gang: Farnell 1114 606 5K only.
    Log, Cermet, Single gang: Farnell 1141 608 to 611 (1K,5K,10K,50K)
    Log, Plastic, Dual gang: Farnell 1141 613 10K only. Note 50mm shaft length, all other 25mm.
    Knobs to fit: 16.2mm diameter Red: RS 259-6935, Black: RS 259-6941'''
    desc = 'Potentiometer, rotary, Vishay Sfernice P11 [Farnell 1141 595] etc.'
    drillMark( f, x, y, 100 )
    drawCircleDotted( f, x, y, 81 )
    compCounts['pot'] += 1
    outlineLabelShifted( f, x, y, 125, 180, 0, -21, 'pot', compRefString('pot', desc) )


def potLabel( f, x, y, label ):
    '''Potentiometer: Rotary, Vishay Sfernice P11, panel or drill pattern.'''
    if( dolabels ):
        r = 162 * 0.5
        lineWidth( f, 2 )
        drawCircle( f, x, y, r )
        drawTextCB( f, x, y-100, label )
        for i in range(0,11):
            t = 225 - ( i * 270 / 10 )
            x1 = r * math.cos( math.radians( t ) ) + x
            y1 = r * math.sin( math.radians( t ) ) + y
            x2 = (r+15) * math.cos( math.radians( t ) ) + x
            y2 = (r+15) * math.sin( math.radians( t ) ) + y
            drawLine( f, x1, y1, x2, y2 )
            if( i == 0 or i == 10 ):
                x2 = (r+25) * math.cos( math.radians( t ) ) + x
                y2 = (r+25) * math.sin( math.radians( t ) ) + y
                drawTextCC( f, x2, y2, '{0}'.format(i) )
    if( domarks ):
        pot( f, x, y )

def fuse( f, x, y ):
    '''Fuse holder: Mains, 5x20mm fuse, panel mounting. Schurter 3101.0110 Farnell 1081 794.'''
    desc = 'Fuse holder, mains, 5x20mm fuse, panel mounting. Schurter 3101.0110 [Farnell 1081 794].'
    drillMark( f, x, y, 130 );
    compCounts['fuse'] += 1
    outlineLabel( f, x, y, 150, 150, 'fuse', compRefString('fuse', desc) )

def fuseLabel( f, x, y, label ):
    '''Fuse holder, panel or drill pattern.'''
    if( dolabels ):
        lineWidth( f, 2 )
        drawCircle( f, x, y, 80 )
        drawTextCB( f, x, y-100, label )
    if( domarks ):
        fuse( f, x, y )

def inlet( f, x, y ):
    '''Inlet, IEC, panel mount. Multicomp JR-101. Farnell 9521 631'''
    desc = 'Inlet: IEC, panel mount. Multicomp JR-101. [Farnell 9521 631].'
    drillRect( f, x, y, 277, 200 )
    drillMark( f, x-200, y, 35 )
    drillMark( f, x+200, y, 35 )
    compCounts['inlet'] += 1
    outlineLabel( f, x, y, 500, 220, 'inlet', compRefString('inlet', desc) )

def inletLabel( f, x, y, label ):
    '''Inlet, IEC, panel or drill pattern.'''
    if( dolabels ):
        lineWidth( f, 2 )
        drawRect( f, x, y, 297, 220 )
        drawTextCB( f, x, y-120, label )
    if( domarks ):
        inlet( f, x, y )

def swmains( f, x, y ):
    '''Mains switch, rocker, illuminated, green. APEM MR210R5LBG Farnell 1082 460.'''
    desc = 'Mains switch: Rocker, illuminated, green. APEM MR210R5LBG [Farnell 1082 460].'
    drillMark( f, x, y, 200 )
    compCounts['swmains'] += 1
    outlineLabel( f, x, y, 240, 240, 'swmains', compRefString('swmains', desc) )

def swmainsLabel( f, x, y, label ):
    '''Mains switch, panel or drill pattern.'''
    if( dolabels ):
        lineWidth( f, 2 )
        drawCircle( f, x, y, 240*0.5 )
        drawTextCB( f, x, y-260*0.5, label )
    if( domarks ):
        swmains( f, x, y )

def jackqt( f, x, y ):
    '''Jack socket, quarter inch, 3 pole, Schurter 4803.2400 Farnell '''
    desc = 'Jack socket: quarter inch, 3 pole, Schurter 4803.2400'
    drillMark( f, x, y, 115 )
    compCounts['jackqt'] += 1
    outlineLabelShifted( f, x, y, 190, 244, 0, -40, 'jack', compRefString('jackqt', desc) )

def jackqtLabel( f, x, y, label ):
    '''Jack socket, quarter inch, 3 pole, panel or drill pattern.'''
    if( dolabels ):
        lineWidth( f, 2 )
        drawCircle( f, x, y, 80 )
        drawTextCB( f, x, y-100, label )
    if( domarks ):
        jackqt( f, x, y )

def screw( f, x, y, mnum, csk=False ):
    '''Metric screw, M<mnum> (string), optionally countersunk.

    To search for screws at Farnell, use these keywords:
           Driver  Head   Size (DiaXLen)  Material   Finish
    screw  pozi    csk    Mdxl            steel      bzp 
           slt     pan    d:1.5,2,2.5,3   nylon      zp
                   cheese   3.5,4                    blk
                          l:6,8,10,12,
                            16,20
    pozi=pozidrive
    slt=slot
    csk=countersunk
    bzp=bright zinc plated
    zp=zinc plated
    blk=black
    '''
    mscrew_clear = {'M1.5':20, 'M2':25, 'M2.5':30, 'M3':35, 'M3.5':40, 'M4':45, 'M5':55, 'M6':65, 'M7':75}
    mscrew_csk = {'M1.5':35, 'M2':40, 'M2.5':47, 'M3':56, 'M3.5':65, 'M4':75, 'M5':92, 'M6':110, 'M7':130}
    if mnum in mscrew_clear:
        d = mscrew_clear[mnum]
        if( csk ):
            s = '{0} csk'.format( mnum )
        else:
            s = '{0}'.format( mnum )
        desc = s + ' screw'
        drillMark( f, x, y, d )
        if csk:
            d = mscrew_csk[mnum]
            drillMark( f, x, y, d, True )
        compCounts['screw'] += 1
        outlineLabel( f, x, y, d+15, d+15, compRefString('screw', desc) )

def screwLabel( f, x, y, mnum, csk=False, label=' ' ):
    '''Metric screw. No label, only drill pattern.'''
    if( domarks ):
        screw( f, x, y, mnum, csk )

def d9pin( f, x, y ):
    '''9 pin D-connector'''
    desc = '9 pin D-connector'
    drillWedgeRounded( f, x, y, 175, 114, 195, 20 )
    drillMark( f, x-124, y, 35 )
    drillMark( f, x+124, y, 35 )
    compCounts['d9pin'] += 1
    outlineLabel( f, x, y, 310, 125, 'd9pin', compRefString('d9pin', desc) )

def d9pinLabel( f, x, y, label=' ' ):
    if( dolabels ):
        lineWidth( f, 2 )
        drawRectRounded( f, x, y, 320, 135,15 )
        drawTextCB( f, x, y-85, label )
    if( domarks ):
        d9pin( f, x, y )



#
# Front panel graphical elements (not associated with single components) ...
#

def grayBoxLabel( f, x, y, w, h, g, label=' ', font='f1', *optdxdy ):
    ''' Draw a gray box with a label if in panel labels mode.'''
    if( dolabels ):
        grayBox( f, x, y, w, h, g )
        # Default label position in in center.
        lx = 0.5 * w
        ly = 0.5 * h
        # Use optional dx and dy to move label.
        noptargs = len( optdxdy )
        if( noptargs == 1 ): # Have dx ...
            if( optdxdy[0] is not None ):
                lx = optdxdy[0]
        elif( noptargs == 2 ): # Have dx and dy ...
            if( optdxdy[0] is not None ):
                lx = optdxdy[0]
            if( optdxdy[1] is not None ):
                ly = optdxdy[1]
        lx += x
        ly += y
        drawTextCC( f, lx, ly, label, font )

def grayBoxLabelTriplet( f, x, y, w, h, t, label=' ', font='f1', *optdxdy ):
    '''Draw a colour box with colour specified by a triplet t.'''
    r, g, b = t;
    print( r, g, b )
    colourBoxLabel( f, x, y, w, h, r, g, b, label, font )

def colourBoxLabel( f, x, y, w, h, r, g, b, label=' ', font='f1', *optdxdy ):
    ''' Draw a colour box with a label if in panel labels mode.'''
    if( dolabels ):
        colourBox( f, x, y, w, h, r, g, b )
        # Default label position in in center.
        lx = 0.5 * w
        ly = 0.5 * h
        # Use optional dx and dy to move label.
        noptargs = len( optdxdy )
        if( noptargs == 1 ): # Have dx ...
            if( optdxdy[0] is not None ):
                lx = optdxdy[0]
        elif( noptargs == 2 ): # Have dx and dy ...
            if( optdxdy[0] is not None ):
                lx = optdxdy[0]
            if( optdxdy[1] is not None ):
                ly = optdxdy[1]
        lx += x
        ly += y
        drawTextCC( f, lx, ly, label, font )

#
# Component lists, panels and pages ...
#

def countComponents( title=' ' ):
    '''Output how many of each defined component we are using.'''
    if( domarks and docomponents ):
        print( 'Component Counts: {0}'.format( title ) )
        for key in sorted( compCounts ):
            print( '{0:>10} : {1}'.format( key, compCounts[key] ) )

def listComponents( title=' ' ):
    '''List each component reference with its description.'''
    if( domarks and docomponents ):
        print( 'Component References: {0}'.format( title ) )
        compList.sort()
        for i in range(0,len(compList)):
            print( addSpacesToColon( stripInnerZeros( compList[i] ) ) )

def resetLists():
    '''Reset component list and counts.'''
    global compCounts
    global typeCounts
    global compList
    
    for key in compCounts:
        compCounts[key] = 0
    for key in typeCounts:
        typeCounts[key] = 0
    compList = []

def beginPanel():
    '''Start defining a panel.'''
    resetLists()

def endPanel( title=' ' ):
    '''Stop defining a panel. Show the component counts and list the components if doing drill marks.'''
    if( curpage == 0 ):
        countComponents( title )
        listComponents( title )

def beginA4landPage( f, pagenum ):
    '''Begin A4 landscape page pagenum.'''
    global curpage
    curpage = pagenum
    xshift = 350 - pagenum * 1700 #2500 #2500 #1700
    beginPage( f )
    shift( f, xshift, 350 )

def endA4landPage( f ):
    '''End an A4 landscape page.'''
    endshift( f )
    endPage( f )
            
#
# If this module is run, draw the components ...
#

if __name__ == '__main__':

    # Open the output file.
    f = createPSFile( 'components.ps' )
    if f is None:
        print( 'Failed to open file components.ps' )
    else:
        print( 'Opened file components.ps.' )

        # Start this output page.
        beginPage( f )
        shift( f, 250, 250 )

        # Grid parameters 1.
        marginx = 50
        marginy = 50
        stepx = 300
        stepy = 300

        # Draw a grid to group the drawings appropriately.
        ncolumns = 6
        nrows = 3
        beginColour( f, 1, 0.8, 0.8 )
        for i in range(0,ncolumns+1):
            x = i * stepx - marginx
            y0 = -marginy
            y1 = nrows * 2 * stepy - marginy
            drawLine( f, x, y0, x, y1 )
        for i in range(0,nrows+1):
            y = i * 2 * stepy - marginy
            x0 = -marginx
            x1 = ncolumns * stepx - marginx
            drawLine( f, x0, y, x1, y )
        endColour( f )

        # Title.
        drawTextC( f, 0.5*(x0+x1), y1+30, 'panelwork.py - Components', 'f2' )

        # Do twice: First drill pattern, then label.
        for i in range(0,2):
            beginPanel()

            # Choose what to draw.
            if i == 0:
                doControlLabels( False )
                doDrillMarks( True )
            else:
                doControlLabels( True )
                doDrillMarks( False )

            # Grid parameters 2.
            orgx = 2 * marginx
            orgy = 2 * marginy + i * stepy
            curx = orgx
            cury = orgy

            # Led
            ledLabel( f, curx, cury, 'Led 3mm' )
            curx += stepx

            # Switch
            sw1Label( f, curx, cury, 'Bot', 'Top', 'Mid' )
            curx += stepx

            # Switch, Mid on right
            sw1Label( f, curx, cury, 'Bot', 'Top', 'Mid', True )
            curx += stepx

            # Switch with Leds on right
            sw1LabelLedsR( f, curx, cury, 'Bot', 'Top', 'Mid' )
            curx += stepx

            # Switch with Leds on left
            sw1LabelLedsL( f, curx, cury, 'Bot', 'Top', 'Mid' )
            curx += stepx

            # Switch sideways
            sw1SidewaysLabel( f, curx, cury, 'Left', 'Right', 'Mid' )
            curx += stepx

            # Next row.
            curx = orgx
            cury += 2* stepy

            # XLR socket.
            xlrLabel( f, curx, cury, 'XLR Socket' )
            curx += stepx

            # Phono socket
            phonoLabel( f, curx, cury, 'Phono' )
            curx += stepx

            # Potentiometer
            potLabel( f, curx, cury, 'Potentiometer' )
            curx += stepx

            # Mains switch
            swmainsLabel( f, curx, cury, 'Mains' )
            curx += stepx

            # Jack Socket
            jackqtLabel( f, curx, cury, 'Jack' )
            curx += stepx

            # Fuse holder
            fuseLabel( f, curx, cury, 'Fuse' )
            curx += stepx

            # Next row.
            curx = orgx
            cury += 2* stepy

            # IEC mains inlet
            inletLabel( f, curx, cury, 'Mains Inlet' )
            curx += 2 * stepx

            # Screw
            screwLabel( f, curx, cury, 'M3' )
            curx += stepx

            # Screw, countersunk
            screwLabel( f, curx, cury, 'M3', True )
            curx += stepx

            # 9 pin D-connector
            d9pinLabel( f, curx, cury, '9 pin D' )
            curx += stepx

            # Show component counts
            endPanel( 'Component Samples.' )

        # End the page.
        endshift( f )
        endPage( f )

        # Close the output file.
        endPSFile( f )
